MODULE Keyboard;

IMPORT
  Kernel, SYSTEM;

CONST
  TimeUnit* = 1000;	(** portable, but VAR on other ports *)	(** timer ticks per second (platform dependent). *)
  SHIFT* = 0;  CTRL* = 1;  ALT* = 2;	(** for KeyState *)

  BufSize = 32;	(* keyboard buffer size *)
  ScrollLock = 0;
  NumLock = 1;
  CapsLock = 2;
  LAlt = 3;
  RAlt = 4;
  LCtrl = 5;
  RCtrl = 6;
  LShift = 7;
  RShift = 8;
  GreyEsc = 9;
  Resetting = 10;
  SetTypematic = 11;
  SendingLEDs = 12;
  MenuShift = 13;
  DeadKey = 0;

  Trace = TRUE;
  On    = TRUE;
  Off   = FALSE;

TYPE
  PollKeyboard* = PROCEDURE (VAR ch: CHAR; VAR keys: SET): BOOLEAN;	(** non-portable *)
  KeyboardPoller = POINTER TO RECORD
    poll: PollKeyboard;
    next: KeyboardPoller
  END;
  KeyTable = POINTER TO ARRAY OF CHAR;

VAR
 (* keyboard state *)
  keyboard: KeyboardPoller;
  buffer: ARRAY 1+BufSize OF CHAR;	(* first byte not used (System.State security) *)
  head, tail, dkey: SHORTINT;
  lastport: LONGINT;
  lastvalue: SYSTEM.BYTE;
  keyval: INTEGER;
  table: LONGINT;
  flags, pollkeys: SET;
  breakproc, timerproc: Kernel.Proc;
  keytable: KeyTable;	(* anchor for keyboard table loaded from file *)
  kpmap: SET;
  kdx, kdy, counter0, counter1: INTEGER;

  VgaFont: ARRAY 1280 OF CHAR;
  layout : BOOLEAN;

PROCEDURE InitFont;
BEGIN
  VgaFont[0000]:=000X; VgaFont[0001]:=000X; VgaFont[0002]:=01EX; VgaFont[0003]:=036X; VgaFont[0004]:=066X; VgaFont[0005]:=0C6X; VgaFont[0006]:=0C6X; VgaFont[0007]:=0FEX;
  VgaFont[0008]:=0C6X; VgaFont[0009]:=0C6X; VgaFont[0010]:=0C6X; VgaFont[0011]:=0C6X; VgaFont[0012]:=000X; VgaFont[0013]:=000X; VgaFont[0014]:=000X; VgaFont[0015]:=000X;
  VgaFont[0016]:=000X; VgaFont[0017]:=000X; VgaFont[0018]:=0FEX; VgaFont[0019]:=062X; VgaFont[0020]:=062X; VgaFont[0021]:=060X; VgaFont[0022]:=07CX; VgaFont[0023]:=066X;
  VgaFont[0024]:=066X; VgaFont[0025]:=066X; VgaFont[0026]:=066X; VgaFont[0027]:=0FCX; VgaFont[0028]:=000X; VgaFont[0029]:=000X; VgaFont[0030]:=000X; VgaFont[0031]:=000X;
  VgaFont[0032]:=000X; VgaFont[0033]:=000X; VgaFont[0034]:=0FCX; VgaFont[0035]:=066X; VgaFont[0036]:=066X; VgaFont[0037]:=066X; VgaFont[0038]:=07CX; VgaFont[0039]:=066X;
  VgaFont[0040]:=066X; VgaFont[0041]:=066X; VgaFont[0042]:=066X; VgaFont[0043]:=0FCX; VgaFont[0044]:=000X; VgaFont[0045]:=000X; VgaFont[0046]:=000X; VgaFont[0047]:=000X;
  VgaFont[0048]:=000X; VgaFont[0049]:=000X; VgaFont[0050]:=0FEX; VgaFont[0051]:=062X; VgaFont[0052]:=062X; VgaFont[0053]:=060X; VgaFont[0054]:=060X; VgaFont[0055]:=060X;
  VgaFont[0056]:=060X; VgaFont[0057]:=060X; VgaFont[0058]:=060X; VgaFont[0059]:=0F0X; VgaFont[0060]:=000X; VgaFont[0061]:=000X; VgaFont[0062]:=000X; VgaFont[0063]:=000X;
  VgaFont[0064]:=000X; VgaFont[0065]:=000X; VgaFont[0066]:=01EX; VgaFont[0067]:=036X; VgaFont[0068]:=066X; VgaFont[0069]:=066X; VgaFont[0070]:=066X; VgaFont[0071]:=066X;
  VgaFont[0072]:=066X; VgaFont[0073]:=066X; VgaFont[0074]:=066X; VgaFont[0075]:=0FFX; VgaFont[0076]:=0C3X; VgaFont[0077]:=081X; VgaFont[0078]:=000X; VgaFont[0079]:=000X;
  VgaFont[0080]:=000X; VgaFont[0081]:=000X; VgaFont[0082]:=0FEX; VgaFont[0083]:=066X; VgaFont[0084]:=062X; VgaFont[0085]:=068X; VgaFont[0086]:=078X; VgaFont[0087]:=068X;
  VgaFont[0088]:=060X; VgaFont[0089]:=062X; VgaFont[0090]:=066X; VgaFont[0091]:=0FEX; VgaFont[0092]:=000X; VgaFont[0093]:=000X; VgaFont[0094]:=000X; VgaFont[0095]:=000X;
  VgaFont[0096]:=000X; VgaFont[0097]:=000X; VgaFont[0098]:=0D6X; VgaFont[0099]:=0D6X; VgaFont[0100]:=054X; VgaFont[0101]:=054X; VgaFont[0102]:=07CX; VgaFont[0103]:=07CX;
  VgaFont[0104]:=054X; VgaFont[0105]:=0D6X; VgaFont[0106]:=0D6X; VgaFont[0107]:=0D6X; VgaFont[0108]:=000X; VgaFont[0109]:=000X; VgaFont[0110]:=000X; VgaFont[0111]:=000X;
  VgaFont[0112]:=000X; VgaFont[0113]:=000X; VgaFont[0114]:=07CX; VgaFont[0115]:=0C6X; VgaFont[0116]:=006X; VgaFont[0117]:=006X; VgaFont[0118]:=03CX; VgaFont[0119]:=006X;
  VgaFont[0120]:=006X; VgaFont[0121]:=006X; VgaFont[0122]:=0C6X; VgaFont[0123]:=07CX; VgaFont[0124]:=000X; VgaFont[0125]:=000X; VgaFont[0126]:=000X; VgaFont[0127]:=000X;
  VgaFont[0128]:=000X; VgaFont[0129]:=000X; VgaFont[0130]:=0C6X; VgaFont[0131]:=0C6X; VgaFont[0132]:=0CEX; VgaFont[0133]:=0CEX; VgaFont[0134]:=0D6X; VgaFont[0135]:=0E6X;
  VgaFont[0136]:=0E6X; VgaFont[0137]:=0C6X; VgaFont[0138]:=0C6X; VgaFont[0139]:=0C6X; VgaFont[0140]:=000X; VgaFont[0141]:=000X; VgaFont[0142]:=000X; VgaFont[0143]:=000X;
  VgaFont[0144]:=038X; VgaFont[0145]:=038X; VgaFont[0146]:=0C6X; VgaFont[0147]:=0C6X; VgaFont[0148]:=0CEX; VgaFont[0149]:=0CEX; VgaFont[0150]:=0D6X; VgaFont[0151]:=0E6X;
  VgaFont[0152]:=0E6X; VgaFont[0153]:=0C6X; VgaFont[0154]:=0C6X; VgaFont[0155]:=0C6X; VgaFont[0156]:=000X; VgaFont[0157]:=000X; VgaFont[0158]:=000X; VgaFont[0159]:=000X;
  VgaFont[0160]:=000X; VgaFont[0161]:=000X; VgaFont[0162]:=0E6X; VgaFont[0163]:=066X; VgaFont[0164]:=06CX; VgaFont[0165]:=06CX; VgaFont[0166]:=078X; VgaFont[0167]:=078X;
  VgaFont[0168]:=06CX; VgaFont[0169]:=06CX; VgaFont[0170]:=066X; VgaFont[0171]:=0E6X; VgaFont[0172]:=000X; VgaFont[0173]:=000X; VgaFont[0174]:=000X; VgaFont[0175]:=000X;
  VgaFont[0176]:=000X; VgaFont[0177]:=000X; VgaFont[0178]:=01EX; VgaFont[0179]:=036X; VgaFont[0180]:=066X; VgaFont[0181]:=0C6X; VgaFont[0182]:=0C6X; VgaFont[0183]:=0C6X;
  VgaFont[0184]:=0C6X; VgaFont[0185]:=0C6X; VgaFont[0186]:=0C6X; VgaFont[0187]:=0C6X; VgaFont[0188]:=000X; VgaFont[0189]:=000X; VgaFont[0190]:=000X; VgaFont[0191]:=000X;
  VgaFont[0192]:=000X; VgaFont[0193]:=000X; VgaFont[0194]:=0C6X; VgaFont[0195]:=0EEX; VgaFont[0196]:=0FEX; VgaFont[0197]:=0FEX; VgaFont[0198]:=0D6X; VgaFont[0199]:=0C6X;
  VgaFont[0200]:=0C6X; VgaFont[0201]:=0C6X; VgaFont[0202]:=0C6X; VgaFont[0203]:=0C6X; VgaFont[0204]:=000X; VgaFont[0205]:=000X; VgaFont[0206]:=000X; VgaFont[0207]:=000X;
  VgaFont[0208]:=000X; VgaFont[0209]:=000X; VgaFont[0210]:=0C6X; VgaFont[0211]:=0C6X; VgaFont[0212]:=0C6X; VgaFont[0213]:=0C6X; VgaFont[0214]:=0FEX; VgaFont[0215]:=0C6X;
  VgaFont[0216]:=0C6X; VgaFont[0217]:=0C6X; VgaFont[0218]:=0C6X; VgaFont[0219]:=0C6X; VgaFont[0220]:=000X; VgaFont[0221]:=000X; VgaFont[0222]:=000X; VgaFont[0223]:=000X;
  VgaFont[0224]:=000X; VgaFont[0225]:=000X; VgaFont[0226]:=07CX; VgaFont[0227]:=0C6X; VgaFont[0228]:=0C6X; VgaFont[0229]:=0C6X; VgaFont[0230]:=0C6X; VgaFont[0231]:=0C6X;
  VgaFont[0232]:=0C6X; VgaFont[0233]:=0C6X; VgaFont[0234]:=0C6X; VgaFont[0235]:=07CX; VgaFont[0236]:=000X; VgaFont[0237]:=000X; VgaFont[0238]:=000X; VgaFont[0239]:=000X;
  VgaFont[0240]:=000X; VgaFont[0241]:=000X; VgaFont[0242]:=0FEX; VgaFont[0243]:=0C6X; VgaFont[0244]:=0C6X; VgaFont[0245]:=0C6X; VgaFont[0246]:=0C6X; VgaFont[0247]:=0C6X;
  VgaFont[0248]:=0C6X; VgaFont[0249]:=0C6X; VgaFont[0250]:=0C6X; VgaFont[0251]:=0C6X; VgaFont[0252]:=000X; VgaFont[0253]:=000X; VgaFont[0254]:=000X; VgaFont[0255]:=000X;
  VgaFont[0256]:=000X; VgaFont[0257]:=000X; VgaFont[0258]:=0FCX; VgaFont[0259]:=066X; VgaFont[0260]:=066X; VgaFont[0261]:=066X; VgaFont[0262]:=07CX; VgaFont[0263]:=060X;
  VgaFont[0264]:=060X; VgaFont[0265]:=060X; VgaFont[0266]:=060X; VgaFont[0267]:=0F0X; VgaFont[0268]:=000X; VgaFont[0269]:=000X; VgaFont[0270]:=000X; VgaFont[0271]:=000X;
  VgaFont[0272]:=000X; VgaFont[0273]:=000X; VgaFont[0274]:=03CX; VgaFont[0275]:=066X; VgaFont[0276]:=0C2X; VgaFont[0277]:=0C0X; VgaFont[0278]:=0C0X; VgaFont[0279]:=0C0X;
  VgaFont[0280]:=0C0X; VgaFont[0281]:=0C2X; VgaFont[0282]:=066X; VgaFont[0283]:=03CX; VgaFont[0284]:=000X; VgaFont[0285]:=000X; VgaFont[0286]:=000X; VgaFont[0287]:=000X;
  VgaFont[0288]:=000X; VgaFont[0289]:=000X; VgaFont[0290]:=07EX; VgaFont[0291]:=05AX; VgaFont[0292]:=018X; VgaFont[0293]:=018X; VgaFont[0294]:=018X; VgaFont[0295]:=018X;
  VgaFont[0296]:=018X; VgaFont[0297]:=018X; VgaFont[0298]:=018X; VgaFont[0299]:=03CX; VgaFont[0300]:=000X; VgaFont[0301]:=000X; VgaFont[0302]:=000X; VgaFont[0303]:=000X;
  VgaFont[0304]:=000X; VgaFont[0305]:=000X; VgaFont[0306]:=0C6X; VgaFont[0307]:=0C6X; VgaFont[0308]:=0C6X; VgaFont[0309]:=0C6X; VgaFont[0310]:=0C6X; VgaFont[0311]:=07EX;
  VgaFont[0312]:=006X; VgaFont[0313]:=006X; VgaFont[0314]:=0C6X; VgaFont[0315]:=07CX; VgaFont[0316]:=000X; VgaFont[0317]:=000X; VgaFont[0318]:=000X; VgaFont[0319]:=000X;
  VgaFont[0320]:=000X; VgaFont[0321]:=03CX; VgaFont[0322]:=018X; VgaFont[0323]:=07EX; VgaFont[0324]:=0DBX; VgaFont[0325]:=0DBX; VgaFont[0326]:=0DBX; VgaFont[0327]:=0DBX;
  VgaFont[0328]:=0DBX; VgaFont[0329]:=07EX; VgaFont[0330]:=018X; VgaFont[0331]:=03CX; VgaFont[0332]:=000X; VgaFont[0333]:=000X; VgaFont[0334]:=000X; VgaFont[0335]:=000X;
  VgaFont[0336]:=000X; VgaFont[0337]:=000X; VgaFont[0338]:=0C6X; VgaFont[0339]:=0C6X; VgaFont[0340]:=06CX; VgaFont[0341]:=07CX; VgaFont[0342]:=038X; VgaFont[0343]:=038X;
  VgaFont[0344]:=07CX; VgaFont[0345]:=06CX; VgaFont[0346]:=0C6X; VgaFont[0347]:=0C6X; VgaFont[0348]:=000X; VgaFont[0349]:=000X; VgaFont[0350]:=000X; VgaFont[0351]:=000X;
  VgaFont[0352]:=000X; VgaFont[0353]:=000X; VgaFont[0354]:=0CCX; VgaFont[0355]:=0CCX; VgaFont[0356]:=0CCX; VgaFont[0357]:=0CCX; VgaFont[0358]:=0CCX; VgaFont[0359]:=0CCX;
  VgaFont[0360]:=0CCX; VgaFont[0361]:=0CCX; VgaFont[0362]:=0CCX; VgaFont[0363]:=0FEX; VgaFont[0364]:=006X; VgaFont[0365]:=006X; VgaFont[0366]:=000X; VgaFont[0367]:=000X;
  VgaFont[0368]:=000X; VgaFont[0369]:=000X; VgaFont[0370]:=0C6X; VgaFont[0371]:=0C6X; VgaFont[0372]:=0C6X; VgaFont[0373]:=0C6X; VgaFont[0374]:=0C6X; VgaFont[0375]:=07EX;
  VgaFont[0376]:=006X; VgaFont[0377]:=006X; VgaFont[0378]:=006X; VgaFont[0379]:=006X; VgaFont[0380]:=000X; VgaFont[0381]:=000X; VgaFont[0382]:=000X; VgaFont[0383]:=000X;
  VgaFont[0384]:=000X; VgaFont[0385]:=000X; VgaFont[0386]:=0DBX; VgaFont[0387]:=0DBX; VgaFont[0388]:=0DBX; VgaFont[0389]:=0DBX; VgaFont[0390]:=0DBX; VgaFont[0391]:=0DBX;
  VgaFont[0392]:=0DBX; VgaFont[0393]:=0DBX; VgaFont[0394]:=0DBX; VgaFont[0395]:=0FFX; VgaFont[0396]:=000X; VgaFont[0397]:=000X; VgaFont[0398]:=000X; VgaFont[0399]:=000X;
  VgaFont[0400]:=000X; VgaFont[0401]:=000X; VgaFont[0402]:=0DBX; VgaFont[0403]:=0DBX; VgaFont[0404]:=0DBX; VgaFont[0405]:=0DBX; VgaFont[0406]:=0DBX; VgaFont[0407]:=0DBX;
  VgaFont[0408]:=0DBX; VgaFont[0409]:=0DBX; VgaFont[0410]:=0DBX; VgaFont[0411]:=0FFX; VgaFont[0412]:=003X; VgaFont[0413]:=003X; VgaFont[0414]:=000X; VgaFont[0415]:=000X;
  VgaFont[0416]:=000X; VgaFont[0417]:=000X; VgaFont[0418]:=0F8X; VgaFont[0419]:=0B0X; VgaFont[0420]:=030X; VgaFont[0421]:=030X; VgaFont[0422]:=03EX; VgaFont[0423]:=033X;
  VgaFont[0424]:=033X; VgaFont[0425]:=033X; VgaFont[0426]:=033X; VgaFont[0427]:=07EX; VgaFont[0428]:=000X; VgaFont[0429]:=000X; VgaFont[0430]:=000X; VgaFont[0431]:=000X;
  VgaFont[0432]:=000X; VgaFont[0433]:=000X; VgaFont[0434]:=0C3X; VgaFont[0435]:=0C3X; VgaFont[0436]:=0C3X; VgaFont[0437]:=0C3X; VgaFont[0438]:=0F3X; VgaFont[0439]:=0DBX;
  VgaFont[0440]:=0DBX; VgaFont[0441]:=0DBX; VgaFont[0442]:=0DBX; VgaFont[0443]:=0F3X; VgaFont[0444]:=000X; VgaFont[0445]:=000X; VgaFont[0446]:=000X; VgaFont[0447]:=000X;
  VgaFont[0448]:=000X; VgaFont[0449]:=000X; VgaFont[0450]:=0F0X; VgaFont[0451]:=060X; VgaFont[0452]:=060X; VgaFont[0453]:=060X; VgaFont[0454]:=07CX; VgaFont[0455]:=066X;
  VgaFont[0456]:=066X; VgaFont[0457]:=066X; VgaFont[0458]:=066X; VgaFont[0459]:=0FCX; VgaFont[0460]:=000X; VgaFont[0461]:=000X; VgaFont[0462]:=000X; VgaFont[0463]:=000X;
  VgaFont[0464]:=000X; VgaFont[0465]:=000X; VgaFont[0466]:=07CX; VgaFont[0467]:=0C6X; VgaFont[0468]:=006X; VgaFont[0469]:=026X; VgaFont[0470]:=03EX; VgaFont[0471]:=026X;
  VgaFont[0472]:=006X; VgaFont[0473]:=006X; VgaFont[0474]:=0C6X; VgaFont[0475]:=07CX; VgaFont[0476]:=000X; VgaFont[0477]:=000X; VgaFont[0478]:=000X; VgaFont[0479]:=000X;
  VgaFont[0480]:=000X; VgaFont[0481]:=000X; VgaFont[0482]:=0CEX; VgaFont[0483]:=0DBX; VgaFont[0484]:=0DBX; VgaFont[0485]:=0DBX; VgaFont[0486]:=0FBX; VgaFont[0487]:=0DBX;
  VgaFont[0488]:=0DBX; VgaFont[0489]:=0DBX; VgaFont[0490]:=0DBX; VgaFont[0491]:=0CEX; VgaFont[0492]:=000X; VgaFont[0493]:=000X; VgaFont[0494]:=000X; VgaFont[0495]:=000X;
  VgaFont[0496]:=000X; VgaFont[0497]:=000X; VgaFont[0498]:=03FX; VgaFont[0499]:=066X; VgaFont[0500]:=066X; VgaFont[0501]:=066X; VgaFont[0502]:=03EX; VgaFont[0503]:=03EX;
  VgaFont[0504]:=066X; VgaFont[0505]:=066X; VgaFont[0506]:=066X; VgaFont[0507]:=0E7X; VgaFont[0508]:=000X; VgaFont[0509]:=000X; VgaFont[0510]:=000X; VgaFont[0511]:=000X;
  VgaFont[0512]:=000X; VgaFont[0513]:=000X; VgaFont[0514]:=000X; VgaFont[0515]:=000X; VgaFont[0516]:=000X; VgaFont[0517]:=078X; VgaFont[0518]:=00CX; VgaFont[0519]:=07CX;
  VgaFont[0520]:=0CCX; VgaFont[0521]:=0CCX; VgaFont[0522]:=0CCX; VgaFont[0523]:=076X; VgaFont[0524]:=000X; VgaFont[0525]:=000X; VgaFont[0526]:=000X; VgaFont[0527]:=000X;
  VgaFont[0528]:=000X; VgaFont[0529]:=002X; VgaFont[0530]:=006X; VgaFont[0531]:=03CX; VgaFont[0532]:=060X; VgaFont[0533]:=060X; VgaFont[0534]:=07CX; VgaFont[0535]:=066X;
  VgaFont[0536]:=066X; VgaFont[0537]:=066X; VgaFont[0538]:=066X; VgaFont[0539]:=03CX; VgaFont[0540]:=000X; VgaFont[0541]:=000X; VgaFont[0542]:=000X; VgaFont[0543]:=000X;
  VgaFont[0544]:=000X; VgaFont[0545]:=000X; VgaFont[0546]:=000X; VgaFont[0547]:=000X; VgaFont[0548]:=000X; VgaFont[0549]:=0FCX; VgaFont[0550]:=066X; VgaFont[0551]:=066X;
  VgaFont[0552]:=07CX; VgaFont[0553]:=066X; VgaFont[0554]:=066X; VgaFont[0555]:=0FCX; VgaFont[0556]:=000X; VgaFont[0557]:=000X; VgaFont[0558]:=000X; VgaFont[0559]:=000X;
  VgaFont[0560]:=000X; VgaFont[0561]:=000X; VgaFont[0562]:=000X; VgaFont[0563]:=000X; VgaFont[0564]:=000X; VgaFont[0565]:=07EX; VgaFont[0566]:=032X; VgaFont[0567]:=032X;
  VgaFont[0568]:=030X; VgaFont[0569]:=030X; VgaFont[0570]:=030X; VgaFont[0571]:=078X; VgaFont[0572]:=000X; VgaFont[0573]:=000X; VgaFont[0574]:=000X; VgaFont[0575]:=000X;
  VgaFont[0576]:=000X; VgaFont[0577]:=000X; VgaFont[0578]:=000X; VgaFont[0579]:=000X; VgaFont[0580]:=000X; VgaFont[0581]:=01EX; VgaFont[0582]:=036X; VgaFont[0583]:=036X;
  VgaFont[0584]:=066X; VgaFont[0585]:=066X; VgaFont[0586]:=066X; VgaFont[0587]:=0FFX; VgaFont[0588]:=0C3X; VgaFont[0589]:=0C3X; VgaFont[0590]:=000X; VgaFont[0591]:=000X;
  VgaFont[0592]:=000X; VgaFont[0593]:=000X; VgaFont[0594]:=000X; VgaFont[0595]:=000X; VgaFont[0596]:=000X; VgaFont[0597]:=07CX; VgaFont[0598]:=0C6X; VgaFont[0599]:=0FEX;
  VgaFont[0600]:=0C0X; VgaFont[0601]:=0C0X; VgaFont[0602]:=0C6X; VgaFont[0603]:=07CX; VgaFont[0604]:=000X; VgaFont[0605]:=000X; VgaFont[0606]:=000X; VgaFont[0607]:=000X;
  VgaFont[0608]:=000X; VgaFont[0609]:=000X; VgaFont[0610]:=000X; VgaFont[0611]:=000X; VgaFont[0612]:=000X; VgaFont[0613]:=0D6X; VgaFont[0614]:=0D6X; VgaFont[0615]:=054X;
  VgaFont[0616]:=07CX; VgaFont[0617]:=054X; VgaFont[0618]:=0D6X; VgaFont[0619]:=0D6X; VgaFont[0620]:=000X; VgaFont[0621]:=000X; VgaFont[0622]:=000X; VgaFont[0623]:=000X;
  VgaFont[0624]:=000X; VgaFont[0625]:=000X; VgaFont[0626]:=000X; VgaFont[0627]:=000X; VgaFont[0628]:=000X; VgaFont[0629]:=03CX; VgaFont[0630]:=066X; VgaFont[0631]:=006X;
  VgaFont[0632]:=00CX; VgaFont[0633]:=006X; VgaFont[0634]:=066X; VgaFont[0635]:=03CX; VgaFont[0636]:=000X; VgaFont[0637]:=000X; VgaFont[0638]:=000X; VgaFont[0639]:=000X;
  VgaFont[0640]:=000X; VgaFont[0641]:=000X; VgaFont[0642]:=000X; VgaFont[0643]:=000X; VgaFont[0644]:=000X; VgaFont[0645]:=0C6X; VgaFont[0646]:=0C6X; VgaFont[0647]:=0CEX;
  VgaFont[0648]:=0D6X; VgaFont[0649]:=0E6X; VgaFont[0650]:=0C6X; VgaFont[0651]:=0C6X; VgaFont[0652]:=000X; VgaFont[0653]:=000X; VgaFont[0654]:=000X; VgaFont[0655]:=000X;
  VgaFont[0656]:=000X; VgaFont[0657]:=000X; VgaFont[0658]:=000X; VgaFont[0659]:=038X; VgaFont[0660]:=038X; VgaFont[0661]:=0C6X; VgaFont[0662]:=0C6X; VgaFont[0663]:=0CEX;
  VgaFont[0664]:=0D6X; VgaFont[0665]:=0E6X; VgaFont[0666]:=0C6X; VgaFont[0667]:=0C6X; VgaFont[0668]:=000X; VgaFont[0669]:=000X; VgaFont[0670]:=000X; VgaFont[0671]:=000X;
  VgaFont[0672]:=000X; VgaFont[0673]:=000X; VgaFont[0674]:=000X; VgaFont[0675]:=000X; VgaFont[0676]:=000X; VgaFont[0677]:=0E6X; VgaFont[0678]:=06CX; VgaFont[0679]:=078X;
  VgaFont[0680]:=078X; VgaFont[0681]:=06CX; VgaFont[0682]:=066X; VgaFont[0683]:=0E6X; VgaFont[0684]:=000X; VgaFont[0685]:=000X; VgaFont[0686]:=000X; VgaFont[0687]:=000X;
  VgaFont[0688]:=000X; VgaFont[0689]:=000X; VgaFont[0690]:=000X; VgaFont[0691]:=000X; VgaFont[0692]:=000X; VgaFont[0693]:=01EX; VgaFont[0694]:=036X; VgaFont[0695]:=066X;
  VgaFont[0696]:=066X; VgaFont[0697]:=066X; VgaFont[0698]:=066X; VgaFont[0699]:=066X; VgaFont[0700]:=000X; VgaFont[0701]:=000X; VgaFont[0702]:=000X; VgaFont[0703]:=000X;
  VgaFont[0704]:=000X; VgaFont[0705]:=000X; VgaFont[0706]:=000X; VgaFont[0707]:=000X; VgaFont[0708]:=000X; VgaFont[0709]:=0C6X; VgaFont[0710]:=0EEX; VgaFont[0711]:=0FEX;
  VgaFont[0712]:=0FEX; VgaFont[0713]:=0D6X; VgaFont[0714]:=0D6X; VgaFont[0715]:=0C6X; VgaFont[0716]:=000X; VgaFont[0717]:=000X; VgaFont[0718]:=000X; VgaFont[0719]:=000X;
  VgaFont[0720]:=000X; VgaFont[0721]:=000X; VgaFont[0722]:=000X; VgaFont[0723]:=000X; VgaFont[0724]:=000X; VgaFont[0725]:=0C6X; VgaFont[0726]:=0C6X; VgaFont[0727]:=0C6X;
  VgaFont[0728]:=0FEX; VgaFont[0729]:=0C6X; VgaFont[0730]:=0C6X; VgaFont[0731]:=0C6X; VgaFont[0732]:=000X; VgaFont[0733]:=000X; VgaFont[0734]:=000X; VgaFont[0735]:=000X;
  VgaFont[0736]:=000X; VgaFont[0737]:=000X; VgaFont[0738]:=000X; VgaFont[0739]:=000X; VgaFont[0740]:=000X; VgaFont[0741]:=07CX; VgaFont[0742]:=0C6X; VgaFont[0743]:=0C6X;
  VgaFont[0744]:=0C6X; VgaFont[0745]:=0C6X; VgaFont[0746]:=0C6X; VgaFont[0747]:=07CX; VgaFont[0748]:=000X; VgaFont[0749]:=000X; VgaFont[0750]:=000X; VgaFont[0751]:=000X;
  VgaFont[0752]:=000X; VgaFont[0753]:=000X; VgaFont[0754]:=000X; VgaFont[0755]:=000X; VgaFont[0756]:=000X; VgaFont[0757]:=0FEX; VgaFont[0758]:=0C6X; VgaFont[0759]:=0C6X;
  VgaFont[0760]:=0C6X; VgaFont[0761]:=0C6X; VgaFont[0762]:=0C6X; VgaFont[0763]:=0C6X; VgaFont[0764]:=000X; VgaFont[0765]:=000X; VgaFont[0766]:=000X; VgaFont[0767]:=000X;

  VgaFont[0768]:=000X; VgaFont[0769]:=000X; VgaFont[0770]:=000X; VgaFont[0771]:=000X; VgaFont[0772]:=000X; VgaFont[0773]:=0DCX; VgaFont[0774]:=066X; VgaFont[0775]:=066X;
  VgaFont[0776]:=066X; VgaFont[0777]:=066X; VgaFont[0778]:=066X; VgaFont[0779]:=07CX; VgaFont[0780]:=060X; VgaFont[0781]:=060X; VgaFont[0782]:=0F0X; VgaFont[0783]:=000X;
  VgaFont[0784]:=000X; VgaFont[0785]:=000X; VgaFont[0786]:=000X; VgaFont[0787]:=000X; VgaFont[0788]:=000X; VgaFont[0789]:=07CX; VgaFont[0790]:=0C6X; VgaFont[0791]:=0C0X;
  VgaFont[0792]:=0C0X; VgaFont[0793]:=0C0X; VgaFont[0794]:=0C6X; VgaFont[0795]:=07CX; VgaFont[0796]:=000X; VgaFont[0797]:=000X; VgaFont[0798]:=000X; VgaFont[0799]:=000X;
  VgaFont[0800]:=000X; VgaFont[0801]:=000X; VgaFont[0802]:=000X; VgaFont[0803]:=000X; VgaFont[0804]:=000X; VgaFont[0805]:=07EX; VgaFont[0806]:=05AX; VgaFont[0807]:=018X;
  VgaFont[0808]:=018X; VgaFont[0809]:=018X; VgaFont[0810]:=018X; VgaFont[0811]:=03CX; VgaFont[0812]:=000X; VgaFont[0813]:=000X; VgaFont[0814]:=000X; VgaFont[0815]:=000X;
  VgaFont[0816]:=000X; VgaFont[0817]:=000X; VgaFont[0818]:=000X; VgaFont[0819]:=000X; VgaFont[0820]:=000X; VgaFont[0821]:=0C6X; VgaFont[0822]:=0C6X; VgaFont[0823]:=0C6X;
  VgaFont[0824]:=0C6X; VgaFont[0825]:=0C6X; VgaFont[0826]:=07EX; VgaFont[0827]:=006X; VgaFont[0828]:=006X; VgaFont[0829]:=0C6X; VgaFont[0830]:=07CX; VgaFont[0831]:=000X;
  VgaFont[0832]:=000X; VgaFont[0833]:=000X; VgaFont[0834]:=000X; VgaFont[0835]:=000X; VgaFont[0836]:=03CX; VgaFont[0837]:=018X; VgaFont[0838]:=07EX; VgaFont[0839]:=0DBX;
  VgaFont[0840]:=0DBX; VgaFont[0841]:=0DBX; VgaFont[0842]:=0DBX; VgaFont[0843]:=07EX; VgaFont[0844]:=018X; VgaFont[0845]:=018X; VgaFont[0846]:=03CX; VgaFont[0847]:=000X;
  VgaFont[0848]:=000X; VgaFont[0849]:=000X; VgaFont[0850]:=000X; VgaFont[0851]:=000X; VgaFont[0852]:=000X; VgaFont[0853]:=0C6X; VgaFont[0854]:=06CX; VgaFont[0855]:=038X;
  VgaFont[0856]:=038X; VgaFont[0857]:=038X; VgaFont[0858]:=06CX; VgaFont[0859]:=0C6X; VgaFont[0860]:=000X; VgaFont[0861]:=000X; VgaFont[0862]:=000X; VgaFont[0863]:=000X;
  VgaFont[0864]:=000X; VgaFont[0865]:=000X; VgaFont[0866]:=000X; VgaFont[0867]:=000X; VgaFont[0868]:=000X; VgaFont[0869]:=0CCX; VgaFont[0870]:=0CCX; VgaFont[0871]:=0CCX;
  VgaFont[0872]:=0CCX; VgaFont[0873]:=0CCX; VgaFont[0874]:=0CCX; VgaFont[0875]:=0FEX; VgaFont[0876]:=006X; VgaFont[0877]:=006X; VgaFont[0878]:=000X; VgaFont[0879]:=000X;
  VgaFont[0880]:=000X; VgaFont[0881]:=000X; VgaFont[0882]:=000X; VgaFont[0883]:=000X; VgaFont[0884]:=000X; VgaFont[0885]:=0C6X; VgaFont[0886]:=0C6X; VgaFont[0887]:=0C6X;
  VgaFont[0888]:=0C6X; VgaFont[0889]:=07EX; VgaFont[0890]:=006X; VgaFont[0891]:=006X; VgaFont[0892]:=000X; VgaFont[0893]:=000X; VgaFont[0894]:=000X; VgaFont[0895]:=000X;
  VgaFont[0896]:=000X; VgaFont[0897]:=000X; VgaFont[0898]:=000X; VgaFont[0899]:=000X; VgaFont[0900]:=000X; VgaFont[0901]:=0D6X; VgaFont[0902]:=0D6X; VgaFont[0903]:=0D6X;
  VgaFont[0904]:=0D6X; VgaFont[0905]:=0D6X; VgaFont[0906]:=0D6X; VgaFont[0907]:=0FEX; VgaFont[0908]:=000X; VgaFont[0909]:=000X; VgaFont[0910]:=000X; VgaFont[0911]:=000X;
  VgaFont[0912]:=000X; VgaFont[0913]:=000X; VgaFont[0914]:=000X; VgaFont[0915]:=000X; VgaFont[0916]:=000X; VgaFont[0917]:=0D6X; VgaFont[0918]:=0D6X; VgaFont[0919]:=0D6X;
  VgaFont[0920]:=0D6X; VgaFont[0921]:=0D6X; VgaFont[0922]:=0D6X; VgaFont[0923]:=0FEX; VgaFont[0924]:=003X; VgaFont[0925]:=003X; VgaFont[0926]:=000X; VgaFont[0927]:=000X;
  VgaFont[0928]:=000X; VgaFont[0929]:=000X; VgaFont[0930]:=000X; VgaFont[0931]:=000X; VgaFont[0932]:=000X; VgaFont[0933]:=0F8X; VgaFont[0934]:=0B0X; VgaFont[0935]:=030X;
  VgaFont[0936]:=03EX; VgaFont[0937]:=033X; VgaFont[0938]:=033X; VgaFont[0939]:=07EX; VgaFont[0940]:=000X; VgaFont[0941]:=000X; VgaFont[0942]:=000X; VgaFont[0943]:=000X;
  VgaFont[0944]:=000X; VgaFont[0945]:=000X; VgaFont[0946]:=000X; VgaFont[0947]:=000X; VgaFont[0948]:=000X; VgaFont[0949]:=0C6X; VgaFont[0950]:=0C6X; VgaFont[0951]:=0C6X;
  VgaFont[0952]:=0F6X; VgaFont[0953]:=0DEX; VgaFont[0954]:=0DEX; VgaFont[0955]:=0F6X; VgaFont[0956]:=000X; VgaFont[0957]:=000X; VgaFont[0958]:=000X; VgaFont[0959]:=000X;
  VgaFont[0960]:=000X; VgaFont[0961]:=000X; VgaFont[0962]:=000X; VgaFont[0963]:=000X; VgaFont[0964]:=000X; VgaFont[0965]:=0F0X; VgaFont[0966]:=060X; VgaFont[0967]:=060X;
  VgaFont[0968]:=07CX; VgaFont[0969]:=066X; VgaFont[0970]:=066X; VgaFont[0971]:=0FCX; VgaFont[0972]:=000X; VgaFont[0973]:=000X; VgaFont[0974]:=000X; VgaFont[0975]:=000X;
  VgaFont[0976]:=000X; VgaFont[0977]:=000X; VgaFont[0978]:=000X; VgaFont[0979]:=000X; VgaFont[0980]:=000X; VgaFont[0981]:=03CX; VgaFont[0982]:=066X; VgaFont[0983]:=006X;
  VgaFont[0984]:=01EX; VgaFont[0985]:=006X; VgaFont[0986]:=066X; VgaFont[0987]:=03CX; VgaFont[0988]:=000X; VgaFont[0989]:=000X; VgaFont[0990]:=000X; VgaFont[0991]:=000X;
  VgaFont[0992]:=000X; VgaFont[0993]:=000X; VgaFont[0994]:=000X; VgaFont[0995]:=000X; VgaFont[0996]:=000X; VgaFont[0997]:=0CEX; VgaFont[0998]:=0DBX; VgaFont[0999]:=0DBX;
  VgaFont[1000]:=0FBX; VgaFont[1001]:=0DBX; VgaFont[1002]:=0DBX; VgaFont[1003]:=0CEX; VgaFont[1004]:=000X; VgaFont[1005]:=000X; VgaFont[1006]:=000X; VgaFont[1007]:=000X;
  VgaFont[1008]:=000X; VgaFont[1009]:=000X; VgaFont[1010]:=000X; VgaFont[1011]:=000X; VgaFont[1012]:=000X; VgaFont[1013]:=07EX; VgaFont[1014]:=0CCX; VgaFont[1015]:=0CCX;
  VgaFont[1016]:=0FCX; VgaFont[1017]:=06CX; VgaFont[1018]:=0CCX; VgaFont[1019]:=0CEX; VgaFont[1020]:=000X; VgaFont[1021]:=000X; VgaFont[1022]:=000X; VgaFont[1023]:=000X;
  VgaFont[1024]:=000X; VgaFont[1025]:=06CX; VgaFont[1026]:=000X; VgaFont[1027]:=0FEX; VgaFont[1028]:=062X; VgaFont[1029]:=060X; VgaFont[1030]:=068X; VgaFont[1031]:=078X;
  VgaFont[1032]:=068X; VgaFont[1033]:=060X; VgaFont[1034]:=062X; VgaFont[1035]:=0FEX; VgaFont[1036]:=000X; VgaFont[1037]:=000X; VgaFont[1038]:=000X; VgaFont[1039]:=000X;
  VgaFont[1040]:=000X; VgaFont[1041]:=000X; VgaFont[1042]:=000X; VgaFont[1043]:=06CX; VgaFont[1044]:=000X; VgaFont[1045]:=07CX; VgaFont[1046]:=0C6X; VgaFont[1047]:=0FEX;
  VgaFont[1048]:=0C0X; VgaFont[1049]:=0C0X; VgaFont[1050]:=0C6X; VgaFont[1051]:=07CX; VgaFont[1052]:=000X; VgaFont[1053]:=000X; VgaFont[1054]:=000X; VgaFont[1055]:=000X;
  VgaFont[1056]:=000X; VgaFont[1057]:=000X; VgaFont[1058]:=000X; VgaFont[1059]:=030X; VgaFont[1060]:=018X; VgaFont[1061]:=00CX; VgaFont[1062]:=006X; VgaFont[1063]:=00CX;
  VgaFont[1064]:=018X; VgaFont[1065]:=030X; VgaFont[1066]:=000X; VgaFont[1067]:=07EX; VgaFont[1068]:=000X; VgaFont[1069]:=000X; VgaFont[1070]:=000X; VgaFont[1071]:=000X;
  VgaFont[1072]:=000X; VgaFont[1073]:=000X; VgaFont[1074]:=000X; VgaFont[1075]:=00CX; VgaFont[1076]:=018X; VgaFont[1077]:=030X; VgaFont[1078]:=060X; VgaFont[1079]:=030X;
  VgaFont[1080]:=018X; VgaFont[1081]:=00CX; VgaFont[1082]:=000X; VgaFont[1083]:=07EX; VgaFont[1084]:=000X; VgaFont[1085]:=000X; VgaFont[1086]:=000X; VgaFont[1087]:=000X;
  VgaFont[1088]:=000X; VgaFont[1089]:=000X; VgaFont[1090]:=00EX; VgaFont[1091]:=01BX; VgaFont[1092]:=01BX; VgaFont[1093]:=018X; VgaFont[1094]:=018X; VgaFont[1095]:=018X;
  VgaFont[1096]:=018X; VgaFont[1097]:=018X; VgaFont[1098]:=018X; VgaFont[1099]:=018X; VgaFont[1100]:=018X; VgaFont[1101]:=018X; VgaFont[1102]:=018X; VgaFont[1103]:=018X;
  VgaFont[1104]:=018X; VgaFont[1105]:=018X; VgaFont[1106]:=018X; VgaFont[1107]:=018X; VgaFont[1108]:=018X; VgaFont[1109]:=018X; VgaFont[1110]:=018X; VgaFont[1111]:=018X;
  VgaFont[1112]:=0D8X; VgaFont[1113]:=0D8X; VgaFont[1114]:=0D8X; VgaFont[1115]:=070X; VgaFont[1116]:=000X; VgaFont[1117]:=000X; VgaFont[1118]:=000X; VgaFont[1119]:=000X;
  VgaFont[1120]:=000X; VgaFont[1121]:=000X; VgaFont[1122]:=000X; VgaFont[1123]:=000X; VgaFont[1124]:=018X; VgaFont[1125]:=018X; VgaFont[1126]:=000X; VgaFont[1127]:=07EX;
  VgaFont[1128]:=000X; VgaFont[1129]:=018X; VgaFont[1130]:=018X; VgaFont[1131]:=000X; VgaFont[1132]:=000X; VgaFont[1133]:=000X; VgaFont[1134]:=000X; VgaFont[1135]:=000X;
  VgaFont[1136]:=000X; VgaFont[1137]:=000X; VgaFont[1138]:=000X; VgaFont[1139]:=000X; VgaFont[1140]:=000X; VgaFont[1141]:=076X; VgaFont[1142]:=0DCX; VgaFont[1143]:=000X;
  VgaFont[1144]:=076X; VgaFont[1145]:=0DCX; VgaFont[1146]:=000X; VgaFont[1147]:=000X; VgaFont[1148]:=000X; VgaFont[1149]:=000X; VgaFont[1150]:=000X; VgaFont[1151]:=000X;
  VgaFont[1152]:=000X; VgaFont[1153]:=038X; VgaFont[1154]:=06CX; VgaFont[1155]:=06CX; VgaFont[1156]:=038X; VgaFont[1157]:=000X; VgaFont[1158]:=000X; VgaFont[1159]:=000X;
  VgaFont[1160]:=000X; VgaFont[1161]:=000X; VgaFont[1162]:=000X; VgaFont[1163]:=000X; VgaFont[1164]:=000X; VgaFont[1165]:=000X; VgaFont[1166]:=000X; VgaFont[1167]:=000X;
  VgaFont[1168]:=000X; VgaFont[1169]:=000X; VgaFont[1170]:=000X; VgaFont[1171]:=000X; VgaFont[1172]:=000X; VgaFont[1173]:=000X; VgaFont[1174]:=000X; VgaFont[1175]:=018X;
  VgaFont[1176]:=018X; VgaFont[1177]:=000X; VgaFont[1178]:=000X; VgaFont[1179]:=000X; VgaFont[1180]:=000X; VgaFont[1181]:=000X; VgaFont[1182]:=000X; VgaFont[1183]:=000X;
  VgaFont[1184]:=000X; VgaFont[1185]:=000X; VgaFont[1186]:=000X; VgaFont[1187]:=000X; VgaFont[1188]:=000X; VgaFont[1189]:=000X; VgaFont[1190]:=000X; VgaFont[1191]:=000X;
  VgaFont[1192]:=018X; VgaFont[1193]:=000X; VgaFont[1194]:=000X; VgaFont[1195]:=000X; VgaFont[1196]:=000X; VgaFont[1197]:=000X; VgaFont[1198]:=000X; VgaFont[1199]:=000X;
  VgaFont[1200]:=000X; VgaFont[1201]:=000X; VgaFont[1202]:=000X; VgaFont[1203]:=001X; VgaFont[1204]:=003X; VgaFont[1205]:=002X; VgaFont[1206]:=006X; VgaFont[1207]:=004X;
  VgaFont[1208]:=0CCX; VgaFont[1209]:=068X; VgaFont[1210]:=038X; VgaFont[1211]:=010X; VgaFont[1212]:=000X; VgaFont[1213]:=000X; VgaFont[1214]:=000X; VgaFont[1215]:=000X;
  VgaFont[1216]:=000X; VgaFont[1217]:=000X; VgaFont[1218]:=000X; VgaFont[1219]:=03CX; VgaFont[1220]:=042X; VgaFont[1221]:=099X; VgaFont[1222]:=0A5X; VgaFont[1223]:=0A1X;
  VgaFont[1224]:=0A5X; VgaFont[1225]:=099X; VgaFont[1226]:=042X; VgaFont[1227]:=03CX; VgaFont[1228]:=000X; VgaFont[1229]:=000X; VgaFont[1230]:=000X; VgaFont[1231]:=000X;
  VgaFont[1232]:=000X; VgaFont[1233]:=070X; VgaFont[1234]:=0D8X; VgaFont[1235]:=030X; VgaFont[1236]:=060X; VgaFont[1237]:=0C8X; VgaFont[1238]:=0F8X; VgaFont[1239]:=000X;
  VgaFont[1240]:=000X; VgaFont[1241]:=000X; VgaFont[1242]:=000X; VgaFont[1243]:=000X; VgaFont[1244]:=000X; VgaFont[1245]:=000X; VgaFont[1246]:=000X; VgaFont[1247]:=000X;
  VgaFont[1248]:=000X; VgaFont[1249]:=000X; VgaFont[1250]:=000X; VgaFont[1251]:=000X; VgaFont[1252]:=07CX; VgaFont[1253]:=07CX; VgaFont[1254]:=07CX; VgaFont[1255]:=07CX;
  VgaFont[1256]:=07CX; VgaFont[1257]:=07CX; VgaFont[1258]:=07CX; VgaFont[1259]:=000X; VgaFont[1260]:=000X; VgaFont[1261]:=000X; VgaFont[1262]:=000X; VgaFont[1263]:=000X;
  VgaFont[1264]:=000X; VgaFont[1265]:=000X; VgaFont[1266]:=000X; VgaFont[1267]:=000X; VgaFont[1268]:=000X; VgaFont[1269]:=000X; VgaFont[1270]:=000X; VgaFont[1271]:=000X;
  VgaFont[1272]:=000X; VgaFont[1273]:=000X; VgaFont[1274]:=042X; VgaFont[1275]:=07EX; VgaFont[1276]:=000X; VgaFont[1277]:=000X; VgaFont[1278]:=000X; VgaFont[1279]:=000X;
END InitFont;

PROCEDURE TableUS(): LONGINT;
CODE {SYSTEM.i386}
  CALL L1
L1:
  POP EAX
  ADD EAX,8
  POP EBP
  RET

  (* alphabet *)
  DB 1EX, "a", "A", 4X
  DB 30X, "b", "B", 4X
  DB 2EX, "c", "C", 4X
  DB 20X, "d", "D", 4X
  DB 12X, "e", "E", 4X
  DB 21X, "f", "F", 4X
  DB 22X, "g", "G", 4X
  DB 23X, "h", "H", 4X
  DB 17X, "i", "I", 4X
  DB 24X, "j", "J", 4X
  DB 25X, "k", "K", 4X
  DB 26X, "l", "L", 4X
  DB 32X, "m", "M", 4X
  DB 31X, "n", "N", 4X
  DB 18X, "o", "O", 4X
  DB 19X, "p", "P", 4X
  DB 10X, "q", "Q", 4X
  DB 13X, "r", "R", 4X
  DB 1FX, "s", "S", 4X
  DB 14X, "t", "T", 4X
  DB 16X, "u", "U", 4X
  DB 2FX, "v", "V", 4X
  DB 11X, "w", "W", 4X
  DB 2DX, "x", "X", 4X
  DB 15X, "y", "Y", 4X
  DB 2CX, "z", "Z", 4X

  (* Oberon accents (LAlt & RAlt) *)
  DB 1EX, "", "", 0CX
  DB 12X, "", 0FFX, 0CX
  DB 18X, "", "", 0CX
  DB 16X, "", "", 0CX
  DB 17X, "", 0FFX, 0CX
  DB 1FX, "", 0FFX, 0CX
  DB 2EX, "", 0FFX, 0CX
  DB 31X, "", 0FFX, 0CX
  DB 1EX, "", "", 14X
  DB 12X, "", 0FFX, 14X
  DB 18X, "", "", 14X
  DB 16X, "", "", 14X
  DB 17X, "", 0FFX, 14X
  DB 1FX, "", 0FFX, 14X
  DB 2EX, "", 0FFX, 14X
  DB 31X, "", 0FFX, 14X

  (* dead keys (LAlt & RAlt) *)
  DB 07X, 0FFX, 1X, 9X
  DB 28X, 2X, 5X, 9X
  DB 29X, 3X, 4X, 9X
  DB 07X, 0FFX, 1X, 11X
  DB 28X, 2X, 5X, 11X
  DB 29X, 3X, 4X, 11X

	  (* following keys *)
  DB 1EX, "", 0FFX, 20X
  DB 12X, "", 0FFX, 20X
  DB 17X, "", 0FFX, 20X
  DB 18X, "", 0FFX, 20X
  DB 16X, "", 0FFX, 20X
  DB 1EX, "", 0FFX, 60X
  DB 12X, "", 0FFX, 60X
  DB 17X, "", 0FFX, 60X
  DB 18X, "", 0FFX, 60X
  DB 16X, "", 0FFX, 60X
  DB 1EX, "", 0FFX, 40X
  DB 12X, "", 0FFX, 40X
  DB 1EX, "", "", 0A4X
  DB 12X, "", 0FFX, 0A0X
  DB 17X, "", 0FFX, 0A0X
  DB 18X, "", "", 0A4X
  DB 16X, "", "", 0A4X
  DB 31X, "", 0FFX, 80X

  (* numbers at top *)
  DB 0BX, "0", ")", 0X
  DB 02X, "1", "!", 0X
  DB 03X, "2", "@", 0X
  DB 04X, "3", "#", 0X
  DB 05X, "4", "$", 0X
  DB 06X, "5", "%", 0X
  DB 07X, "6", "^", 0X
  DB 08X, "7", "&", 0X
  DB 09X, "8", "*", 0X
  DB 0AX, "9", "(", 0X

  (* symbol keys *)
  DB 28X, "'", 22X, 0X
  DB 33X, ",", "<", 0X
  DB 0CX, "-", "_", 0X
  DB 34X, ".", ">", 0X
  DB 35X, "/", "?", 0X
  DB 27X, ";", ":", 0X
  DB 0DX, "=", "+", 0X
  DB 1AX, "[", "{", 0X
  DB 2BX, "\", "|", 0X
  DB 1BX, "]", "}", 0X
  DB 29X, "`", "~", 0X

	  (* control keys *)
  DB 0EX, 7FX, 7FX, 0X	(* backspace *)
  DB 0FX, 09X, 09X, 0X	(* tab *)
  DB 1CX, 0DX, 0DX, 0X	(* enter *)
  DB 39X, 20X, 20X, 0X	(* space *)
  DB 01X, 1BX, 1BX, 0X	(* esc *)
	  (* keypad *)
  DB 04FX, 0A9X, "1" , 2X	(* end/1 *)
  DB 050X, 0C2X, "2" , 2X	(* down/2 *)
  DB 051X, 0A3X, "3" , 2X	(* pgdn/3 *)
  DB 04BX, 0C4X, "4" , 2X	(* left/4 *)
  DB 04CX, 0FFX, "5" , 2X	(* center/5 *)
  DB 04DX, 0C3X, "6" , 2X	(* right/6 *)
  DB 047X, 0A8X, "7" , 2X	(* home/7 *)
  DB 048X, 0C1X, "8" , 2X	(* up/8 *)
  DB 049X, 0A2X, "9" , 2X	(* pgup/9 *)
  DB 052X, 0A0X, "0" , 2X	(* insert/0 *)
  DB 053X, 0A1X, 2EX , 2X	(* del/. *)
	  (* grey keys *)
  DB 04AX, "-" , "-" , 0X	(* grey - *)
  DB 04EX, "+" , "+" , 0X	(* grey + *)
  DB 0B5X, "/" , "/" , 0X	(* grey / *)
  DB 037X, "*" , "*" , 0X	(* grey * *)
  DB 0D0X, 0C2X, 0C2X, 0X	(* grey down *)
  DB 0CBX, 0C4X, 0C4X, 0X	(* grey left *)
  DB 0CDX, 0C3X, 0C3X, 0X	(* grey right *)
  DB 0C8X, 0C1X, 0C1X, 0X	(* grey up *)
  DB 09CX, 00DX, 00DX, 0X	(* grey enter *)
  DB 0D2X, 0A0X, 0A0X, 0X	(* grey ins *)
  DB 0D3X, 0A1X, 0A1X, 0X	(* grey del *)
  DB 0C9X, 0A2X, 0A2X, 0X	(* grey pgup *)
  DB 0D1X, 0A3X, 0A3X, 0X	(* grey pgdn *)
  DB 0C7X, 0A8X, 0A8X, 0X	(* grey home *)
  DB 0CFX, 0A9X, 0A9X, 0X	(* grey end *)
	  (* function keys *)
  DB 03BX, 0A4X, 0FFX, 0X	(* F1 *)
  DB 03CX, 0A5X, 0FFX, 0X	(* F2 *)
  DB 03DX, 01BX, 0FFX, 0X	(* F3 *)
  DB 03EX, 0A7X, 0FFX, 0X	(* F4 *)
  DB 03FX, 0F5X, 0FFX, 0X	(* F5 *)
  DB 040X, 0F6X, 0FFX, 0X	(* F6 *)
  DB 041X, 0F7X, 0FFX, 0X	(* F7 *)
  DB 042X, 0F8X, 0FFX, 0X	(* F8 *)
  DB 043X, 0F9X, 0FFX, 0X	(* F9 *)
  DB 044X, 0FAX, 0FFX, 0X	(* F10 *)
  DB 057X, 0FBX, 0FFX, 0X	(* F11 *)
  DB 058X, 0FCX, 0FFX, 0X	(* F12 *)
  DB 0FFX
END TableUS;

PROCEDURE CheckLayout;
BEGIN
  IF ({LCtrl,LShift} / flags = {})
  OR ({RCtrl,RShift} / flags = {})
  OR ({LCtrl,RShift} / flags = {})
  OR ({RCtrl,LShift} / flags = {})
  THEN
    layout := ~layout;
  END;
END CheckLayout;

PROCEDURE Swap(k: INTEGER):INTEGER;
BEGIN

  IF (CapsLock IN flags) THEN
    IF ((LShift IN flags) OR (RShift IN flags)) THEN
      CASE CHR(k) OF
        | "{" : k:= 229;
        | "}" : k:= 234;
        | ":" : k:= 166;
        | '"' : k:= 237;
        | "<" : k:= 161;
        | ">" : k:= 238;
        | "~" : k:= 241;
      ELSE
      END;
    ELSE
      CASE CHR(k) OF
        | "[" : k:= 149;
        | "]" : k:= 154;
        | ";" : k:= 134;
        | "'" : k:= 157;
        | "," : k:= 129;
        | "." : k:= 158;
        | "`" : k:= 240;
      ELSE
      END;
    END;
  END;

  CASE CHR(k) OF
    | "Q" : k:= 137;
    | "W" : k:= 150;
    | "E" : k:= 147;
    | "R" : k:= 138;
    | "T" : k:= 133;
    | "Y" : k:= 141;
    | "U" : k:= 131;
    | "I" : k:= 152;
    | "O" : k:= 153;
    | "P" : k:= 135;
    | "{" : k:= 149;
    | "}" : k:= 154;
    | "A" : k:= 148;
    | "S" : k:= 155;
    | "D" : k:= 130;
    | "F" : k:= 128;
    | "G" : k:= 143;
    | "H" : k:= 144;
    | "J" : k:= 142;
    | "K" : k:= 139;
    | "L" : k:= 132;
    | ":" : k:= 134;
    | '"' : k:= 157;
    | "Z" : k:= 159;
    | "X" : k:= 151;
    | "C" : k:= 145;
    | "V" : k:= 140;
    | "B" : k:= 136;
    | "N" : k:= 146;
    | "M" : k:= 156;
    | "<" : k:= 129;
    | ">" : k:= 158;
    | "~" : k:= 240;

  ELSE
  END;

  CASE CHR(k) OF
    | "q" : k:= 169;
    | "w" : k:= 230;
    | "e" : k:= 227;
    | "r" : k:= 170;
    | "t" : k:= 165;
    | "y" : k:= 173;
    | "u" : k:= 163;
    | "i" : k:= 232;
    | "o" : k:= 233;
    | "p" : k:= 167;
    | "[" : k:= 229;
    | "]" : k:= 234;
    | "a" : k:= 228;
    | "s" : k:= 235;
    | "d" : k:= 162;
    | "f" : k:= 160;
    | "g" : k:= 175;
    | "h" : k:= 224;
    | "j" : k:= 174;
    | "k" : k:= 171;
    | "l" : k:= 164;
    | ";" : k:= 166;
    | "'" : k:= 237;
    | "z" : k:= 239;
    | "x" : k:= 231;
    | "c" : k:= 225;
    | "v" : k:= 172;
    | "b" : k:= 168;
    | "n" : k:= 226;
    | "m" : k:= 236;
    | "," : k:= 161;
    | "." : k:= 238;
    | "`" : k:= 241;
  ELSE
  END;

  RETURN k;
END Swap;

PROCEDURE Translate(flags: SET;  c: CHAR): INTEGER;
CONST
  Alt = {LAlt, RAlt};
  Ctrl = {LCtrl, RCtrl};
  Shift = {LShift, RShift};
VAR
  a: LONGINT;
  s1: CHAR;
  s: SET;
  k: INTEGER;
  dkn: SHORTINT;
BEGIN
  IF (c = 46X) & (flags * Ctrl # {}) THEN RETURN -2 END;	(* Ctrl-Break - break *)
  IF (c = 44X) & (flags * Ctrl # {}) THEN RETURN 0FFH END;	(* Ctrl-F10 - exit *)
  IF (c = 53X) & (flags * Ctrl # {}) & (flags * Alt # {}) THEN
    RETURN 0FFH
  END;	(* Ctrl-Alt-Del - exit *)
  IF GreyEsc IN flags THEN c := CHR(ORD(c)+80H) END;
  a := table;
  LOOP
    SYSTEM.GET(a, s1);
    IF s1 = 0FFX THEN	(* end of table, unmapped key *)
      k := -1;
      dkey := 0;
      EXIT
    ELSIF s1 = c THEN	(* found scan code in table *)
      SYSTEM.GET(a+3, SYSTEM.VAL(CHAR, s));	(* flags from table *)
      dkn := SHORT(SHORT(SYSTEM.VAL(LONGINT, SYSTEM.LSH(s * {5..7}, -5))));
      s := s * {DeadKey, NumLock, CapsLock, LAlt, RAlt, LCtrl, RCtrl};
      k := 0;
      IF ((s * Alt = flags * Alt) OR (NumLock IN s)) & (dkn = dkey) THEN	(* Alt & dead keys match exactly *)

	IF flags * Shift # {} THEN INCL(s, LShift) END;
	IF (CapsLock IN s) & (CapsLock IN flags) THEN s := s / {LShift} END;
	IF NumLock IN s THEN
	  IF flags * Alt # {} THEN
            INCL(s, LShift)
	  ELSIF NumLock IN flags THEN
            s := s / {LShift}
	  END
	END;

	IF LShift IN s THEN
          SYSTEM.GET(a+2, SYSTEM.VAL(CHAR, k))
	ELSE
          SYSTEM.GET(a+1, SYSTEM.VAL(CHAR, k))
	END;

	IF (DeadKey IN s) & (k <= 7) THEN
	  dkey := SHORT(k);
          k := -1
	ELSIF k = 0FFH THEN
	  k := -1;
          dkey := 0
	ELSE

          IF flags * Ctrl # {} THEN
            IF ((k >= 64) & (k <= 95)) OR ((k >= 97) & (k <= 122)) THEN
	      k := SHORT(SYSTEM.VAL(LONGINT, SYSTEM.VAL(SET, k) * {0..4}))	(* control *)
            ELSIF k = 13 THEN	(* Ctrl-Enter *)
	      k := 10
            END
          END;

          IF flags * Alt # {} THEN	(* Alt-keypad *)
            IF (k >= ORD("0")) & (k <= ORD("9")) & (NumLock IN s) THEN	(* keypad num *)
	      IF keyval = -1 THEN
                keyval := k-ORD("0")
	      ELSE
                keyval := (10*keyval + (k-ORD("0"))) MOD 1000
	      END;
	      k := -1
            END
          END;

          dkey := 0	(* reset dead key state *)
	END;
	EXIT
      END
    END;
    INC(a, 4)
  END; (* LOOP *)

  IF ~layout THEN
    k:=Swap(k);
  END;

  RETURN k
END Translate;

PROCEDURE Wait;
VAR
  t: Kernel.MilliTimer;
  s: SET;
BEGIN
  Kernel.SetTimer(t, Kernel.TimeUnit DIV 50);	(* wait up to 17 ms *)
  REPEAT
    SYSTEM.PORTIN(64H, SYSTEM.VAL(CHAR, s))
  UNTIL ~(1 IN s) OR Kernel.Expired(t)
END Wait;

PROCEDURE SendByte(port: LONGINT;  value: SYSTEM.BYTE);
BEGIN
  Wait;
  SYSTEM.PORTOUT(port, SYSTEM.VAL(CHAR, value));
  lastport := port;
  lastvalue := value
END SendByte;

PROCEDURE ShiftKey(left, right: SHORTINT;  in: BOOLEAN);
BEGIN
  IF in THEN
    IF GreyEsc IN flags THEN
      INCL(flags, right)
    ELSE
      INCL(flags, left)
    END
  ELSE
    IF GreyEsc IN flags THEN
      EXCL(flags, right)
    ELSE
      EXCL(flags, left)
    END
  END
END ShiftKey;

PROCEDURE LedKey(VAR flags: SET;  lock: SHORTINT;  c: CHAR; VAR k: INTEGER);
BEGIN
  IF flags * {LAlt, RAlt, LCtrl, RCtrl, LShift, RShift} = {} THEN
    flags := flags / {lock}
  ELSE
    k := Translate(flags, c)
  END
END LedKey;

PROCEDURE -DisableInterrupts(): SET;
CODE {SYSTEM.i386}
  PUSHFD
  POP EAX
  CLI
END DisableInterrupts;

PROCEDURE -RestoreInterrupts(state: SET);
CODE {SYSTEM.i386}
  POPFD
  STI
END RestoreInterrupts;

PROCEDURE -CS(): LONGINT;
CODE {SYSTEM.i386}
  XOR EAX, EAX
  MOV AX, CS
END CS;

PROCEDURE FntPlane(Flag: BOOLEAN);
BEGIN
  SYSTEM.CLI;
  IF Flag THEN
    SYSTEM.PORTOUT(03C4H, SYSTEM.VAL(INTEGER, 0402H));
    SYSTEM.PORTOUT(03C4H, SYSTEM.VAL(INTEGER, 0704H));
    SYSTEM.PORTOUT(03CEH, SYSTEM.VAL(INTEGER, 0204H));
    SYSTEM.PORTOUT(03CEH, SYSTEM.VAL(INTEGER, 0005H));
    SYSTEM.PORTOUT(03CEH, SYSTEM.VAL(INTEGER, 0406H));
  ELSE
    SYSTEM.PORTOUT(03C4H, SYSTEM.VAL(INTEGER, 0302H));
    SYSTEM.PORTOUT(03C4H, SYSTEM.VAL(INTEGER, 0304H));
    SYSTEM.PORTOUT(03CEH, SYSTEM.VAL(INTEGER, 0004H));
    SYSTEM.PORTOUT(03CEH, SYSTEM.VAL(INTEGER, 1005H));
    SYSTEM.PORTOUT(03CEH, SYSTEM.VAL(INTEGER, 0E06H));
  END;
  SYSTEM.STI;
END FntPlane;

PROCEDURE MapScanCode(c: CHAR): INTEGER;
VAR
  k: INTEGER;
  oldleds, state: SET;
BEGIN
  SendByte(64H, 0ADX);
  Wait;
  k := -1;
  oldleds := flags * {ScrollLock, NumLock, CapsLock};

  IF c = 0X THEN

  ELSIF c = 0FAX THEN
    IF Resetting IN flags THEN
      EXCL(flags, Resetting);
      INCL(flags, SendingLEDs);
      SendByte(60H, 0EDX)	(* set keyboard LEDs *)
    ELSIF SendingLEDs IN flags THEN
      SendByte(60H, SYSTEM.VAL(CHAR, oldleds));
      EXCL(flags, SendingLEDs)
    ELSIF SetTypematic IN flags THEN
      EXCL(flags, SetTypematic);
      INCL(flags, Resetting);
      SendByte(60H, 020X)	(* 30Hz, 500 ms *)
    ELSE (* assume ack was for something else *)
    END
  ELSIF c = 0FEX THEN	(* keyboard resend *)
    SendByte(lastport, lastvalue)
  ELSIF c = 038X THEN	(* Alt make *)
    ShiftKey(LAlt, RAlt, TRUE)
  ELSIF c = 01DX THEN	(* Ctrl make *)
    ShiftKey(LCtrl, RCtrl, TRUE)
  ELSIF c = 02AX THEN	(* LShift make *)
    INCL(flags, LShift)
  ELSIF c = 036X THEN	(* RShift make *)
    INCL(flags, RShift)
  ELSIF c = 05DX THEN	(* menu make *)
    INCL(flags, MenuShift)
  ELSIF c = 03AX THEN	(* Caps make *)
    LedKey(flags, CapsLock, c, k)
  ELSIF c = 046X THEN	(* Scroll make *)
    LedKey(flags, ScrollLock, c, k);
    state := DisableInterrupts();
    IF ScrollLock IN flags THEN
      IF Kernel.timer = NIL THEN Kernel.timer := timerproc END
    ELSE
      IF Kernel.timer = timerproc THEN Kernel.timer := NIL END
    END;
    RestoreInterrupts(state)
  ELSIF c = 045X THEN	(* Num make *)
    LedKey(flags, NumLock, c, k)
  ELSIF c = 0B8X THEN	(* Alt break *)
    ShiftKey(LAlt, RAlt, FALSE);
    IF (keyval >= 0) & (keyval < 255) THEN k := keyval END;	(* exclude 255 - reboot *)
    keyval := -1
  ELSIF c = 09DX THEN	(* Ctrl break *)
    ShiftKey(LCtrl, RCtrl, FALSE)
  ELSIF c = 0AAX THEN	(* LShift break *)
    EXCL(flags, LShift)
  ELSIF c = 0B6X THEN	(* RShift break *)
    EXCL(flags, RShift)
  ELSIF c = 0DDX THEN	(* menu break *)
    EXCL(flags, MenuShift)
  ELSIF (flags * {ScrollLock, GreyEsc} = {ScrollLock}) & (c >= 47X) & (c <= 53X) & (c # 4AX) & (c # 4EX) THEN	(* key mouse *)
    INCL(kpmap, ORD(c)-47H)
  ELSIF c < 080X THEN	(* Other make *)
    k := Translate(flags, c)
  ELSIF (flags * {ScrollLock, GreyEsc} = {ScrollLock}) & (c >= 0C7X) & (c <= 0D3X) THEN	(* key mouse *)
    EXCL(kpmap, ORD(c)-0C7H)
  ELSE

  END;

  IF c = 0E0X THEN
    INCL(flags, GreyEsc)
  ELSE
    EXCL(flags, GreyEsc)
  END;

  IF flags * {ScrollLock, NumLock, CapsLock} # oldleds THEN
    INCL(flags, SendingLEDs);
    SendByte(60H, 0EDX)	(* set keyboard LEDs *)
  END;

  CheckLayout;

  SendByte(64H, 0AEX);	(* enable keyboard *)

  RETURN k
END MapScanCode;

PROCEDURE KeyboardInterrupt;
VAR
  m: SET;
  c: CHAR;
  k: INTEGER;
  fp, esp, tmp, cs: LONGINT;
  i: SHORTINT;
BEGIN

  SYSTEM.PORTIN(060H, c);
  SYSTEM.PORTIN(061H, SYSTEM.VAL(CHAR, m));
  INCL(m, 7);
  SYSTEM.PORTOUT(061H, SYSTEM.VAL(CHAR, m));
  EXCL(m, 7);
  SYSTEM.PORTOUT(061H, SYSTEM.VAL(CHAR, m));
  SYSTEM.STI();

  k := MapScanCode(c);

  IF k = -2 THEN	(* break *)
    head := 0;
    tail := 0;	(* clear buffer *)
    IF ~Kernel.break THEN	(* first try: soft break *)
      Kernel.break := TRUE
    ELSIF ~Kernel.inGC THEN	(* second try: do hard break *)
      Kernel.break := FALSE;	(* cancel other break *)
      SYSTEM.GETREG(5, fp);	(* EBP *)
      SYSTEM.GET(fp+52, tmp);	(* get CS'' *)
      cs := CS();
      IF tmp MOD 4 # cs MOD 4 THEN
        SYSTEM.GET(fp+48, tmp);
        SYSTEM.PUT(fp+48, breakproc);
        SYSTEM.GET(fp+60, esp);
        DEC(esp, 4);
        SYSTEM.PUT(fp+60, esp);
        SYSTEM.PUT(esp, tmp)
      ELSE
        SYSTEM.PUT(fp+48, breakproc)
      END
    END
  ELSIF k >= 0 THEN
    i := (tail+1) MOD BufSize;
    IF i # head THEN
      buffer[1+tail] := CHR(k);
      tail := i
    END
  END;

END KeyboardInterrupt;

PROCEDURE Configure*(config, value: ARRAY OF CHAR);	(** non-portable *)
VAR
  kt: KeyTable;
  state: SET;
  i: SHORTINT;
BEGIN

  IF Trace THEN
    Kernel.WriteString("Input: ");
    Kernel.WriteString(config);
    Kernel.WriteChar("=");
    Kernel.WriteString(value);
    Kernel.WriteLn
  END;

  IF config = "Keyboard" THEN
    i := 0; WHILE (value[i] # 0X) & (value[i] # ".") DO INC(i) END;
    kt := NIL;
    state := DisableInterrupts();
    IF kt = NIL THEN
      table := TableUS()
    ELSE
      keytable := kt;
      table := SYSTEM.ADR(kt[0])
    END;
    dkey := 0;
    RestoreInterrupts(state)
  ELSE
    IF Trace THEN
      Kernel.WriteString("Unknown setting");
      Kernel.WriteLn
    END
  END

END Configure;

PROCEDURE Setting(name: ARRAY OF CHAR);
VAR
  s: ARRAY 32 OF CHAR;
BEGIN
  Kernel.GetConfig(name, s);
  Configure(name, s)
END Setting;

PROCEDURE InitKeyboard;
VAR
  s: SET;
  c: CHAR;
  i: SHORTINT;
  j: INTEGER;
  addr: LONGINT;
  k: ARRAY 8 OF CHAR;
BEGIN

  head := 0;
  tail := 0;
  keyval := -1;
  buffer[0] := 0X;
  Setting("Keyboard");
  flags := {};

  Kernel.InstallIP(KeyboardInterrupt, Kernel.IRQ+1);

  i := 8;
  LOOP
    SYSTEM.PORTIN(64H, SYSTEM.VAL(CHAR, s));
    IF ~(0 IN s) OR (i = 0) THEN EXIT END;
    SYSTEM.PORTIN(60H, c);
    SYSTEM.PORTIN(61H, SYSTEM.VAL(CHAR, s));
    INCL(s, 7);
    SYSTEM.PORTOUT(61H, SYSTEM.VAL(CHAR, s));
    EXCL(s, 7);
    SYSTEM.PORTOUT(61H, SYSTEM.VAL(CHAR, s));
    DEC(i)
  END;

  InitFont;
  FntPlane(On);
  j:=0;
  addr:=0A0000H+128*32;
  WHILE j<1280 DO
    IF j=768 THEN addr:=0A0000H+200*32; END;
    SYSTEM.PUT(addr+j, VgaFont[j]);
    IF (((j+1) MOD 16)=0) THEN INC(addr,16); END;
    INC(j);
  END;
  FntPlane(Off);

  flags := {SetTypematic};
  Kernel.GetConfig("NumLock", k);
  IF k[0] = "1" THEN INCL(flags, NumLock) END;
  SendByte(60H, 0F3X)

END InitKeyboard;

PROCEDURE Read*(VAR ch: CHAR);
VAR
  state: SET;
BEGIN
  REPEAT UNTIL (Available() # 0) OR Kernel.break;
  IF Kernel.break THEN
    Kernel.break := FALSE;
    SYSTEM.HALT(13)
  END;
  state := DisableInterrupts();
  ch := buffer[1+head];
  head := (head+1) MOD BufSize;
  RestoreInterrupts(state)
END Read;

PROCEDURE *UnsafeBreak;
VAR
  note1, note2, note3: ARRAY 32 OF CHAR;
BEGIN
  note1 := "Warning: Interrupting a module";
  note2 := "may invalidate its invariants";
  note3 := "and make it unstable.";
  SYSTEM.HALT(13)
END UnsafeBreak;

PROCEDURE *Timer;
VAR
  i: INTEGER;
BEGIN
  IF counter1 = TimeUnit DIV 100 THEN
    counter1 := 0;

    IF kpmap * {0..2, 4, 6, 8..10} = {} THEN
      counter0 := 0
    ELSIF counter0 < 100 THEN
      INC(counter0)
    END;

    i := counter0 DIV (100 DIV 12)+1;
    IF kpmap * {0,4,8} # {} THEN DEC(kdx, i) END;
    IF kpmap * {0,1,2} # {} THEN DEC(kdy, i) END;
    IF kpmap * {2,6,10} # {} THEN INC(kdx, i) END;
    IF kpmap * {8,9,10} # {} THEN INC(kdy, i) END

  ELSE
    INC(counter1)
  END
END Timer;

PROCEDURE Available*() : INTEGER;
VAR
  state, keys: SET;
  p: KeyboardPoller;
  x: INTEGER;
  ch: CHAR;
  i: SHORTINT;
BEGIN
  p := keyboard;
  pollkeys := {};
  WHILE p # NIL DO
    IF p.poll(ch, keys) THEN
      state := DisableInterrupts();
      i := (tail+1) MOD BufSize;
      IF i # head THEN
        buffer[1+tail] := ch;
        tail := i
      END;
      RestoreInterrupts(state)
    END;
    pollkeys := pollkeys + keys;
    p := p.next
  END;
  state := DisableInterrupts();
  x := (tail-head) MOD BufSize;
  RestoreInterrupts(state);
  RETURN x
END Available;

PROCEDURE Test;
VAR
  ch: CHAR;
  ol: BOOLEAN;
  X,Y: LONGINT;
BEGIN

  Kernel.Cls;

  LOOP
    IF Available() > 0 THEN
      Read(ch);
      Kernel.WriteChar(ch);
    END;

    IF ol#layout THEN
      Kernel.CursorHide;
      IF layout THEN
        Kernel.GetPosition(X,Y);
        Kernel.SetPosition(77,0);
        Kernel.WriteString('Eng');
        Kernel.SetPosition(X,Y);
      ELSE
        Kernel.GetPosition(X,Y);
        Kernel.SetPosition(77,0);
        Kernel.WriteString('Rus');
        Kernel.SetPosition(X,Y);
      END;
      Kernel.CursorShow;
    END;
    ol:=layout;

  END;
END Test;

BEGIN

  keyboard := NIL;
  pollkeys := {};

  timerproc := Timer;
  kpmap := {};
  kdx := 0;
  kdy := 0;
  counter0 := 0;

  layout := TRUE;

  InitKeyboard;
  Test;

END Keyboard.

